Require RAMs_example1.
Require RAMs_to_Linda.
Require Linda.
Require Relations_2.

Definition example1_initial_agent :=
 (agent_of_RAM_and_configuration example1_RAM example1_initial_state).

(* The final state up to process equivalence (i.e. Dead process *)
(* elimination, Parallel lifting to AParallel and commutativity *)
(* and associativity of Parallel).                              *)
Definition example1_final_agent :=
 (agent_of_RAM_and_configuration example1_RAM example1_final_state).

(* The reached final state *)
Definition example1_final_agent' :=
 (AParallel
   (AParallel
     (AParallel
       (AParallel (Program Dead)
         (AParallel (AParallel (Program Dead) (Program Dead))
           (AParallel (AParallel (Program Dead) (Program Dead))
             (AParallel (AParallel (Program Dead) (Program Dead))
               (AParallel (AParallel (Message (6)) (Program Dead))
                 (Program
                   (Action (BangIn (3))
                     (IfThenElseAction (In (1))
                       (Action (Out (4)) Dead)
                       (Action (Out (6)) Dead)))))))))
       (AParallel
         (AParallel (Message (0))
           (AParallel (Program Dead) (Program Dead)))
         (AParallel
           (AParallel (Message (0))
             (AParallel (Program Dead) (Program Dead)))
           (AParallel
             (AParallel (Message (0))
               (AParallel (Program Dead) (Program Dead)))
             (Program
               (Action (BangIn (4))
                 (Action (Out (0)) (Action (Out (5)) Dead))))))))
     (AParallel (AParallel (Program Dead) (Program Dead))
       (AParallel (AParallel (Program Dead) (Program Dead))
         (AParallel (AParallel (Program Dead) (Program Dead))
           (Program
             (Action (BangIn (5))
               (IfThenElseAction (In (2)) (Action (Out (6)) Dead)
                 (Action (Out (3)) Dead))))))))
   (AParallel (Program Dead)
     (AParallel (Program Dead)
       (AParallel
         (AParallel (Program Dead)
           (AParallel (Program Dead)
             (AParallel (Program Dead) (Program Dead))))
         (AParallel
           (AParallel (Message (0))
             (AParallel (Message (0)) (Program Dead)))
           (Program Dead)))))).


(*********************************************)
(* A tactic to partially normalize an agent, *)
(* lifting every Parallel to a AParallel.    *)
(*********************************************)

Tactic Definition GoLeft :=
 EApply SAParallelPL;
  [ Unfold not;
    Induction 1;
    Intros;
    Discriminate
  | Idtac (* The only goal left *)
  ].

Tactic Definition GoRight :=
 EApply SAParallelPR;
  [ Unfold not;
    Induction 1;
    Intros;
    Discriminate
  | Idtac (* The only goal left *)
  ].

Tactic Definition GoDeep :=
 Apply SParallel Orelse (GoLeft ; GoDeep) Orelse (GoRight ; GoDeep).

(* Parallel2AParallel is no longer used. *)
Tactic Definition Parallel2AParallel :=
 Repeat
  (EApply Rstar_n; 
    [ Unfold step;
      Left;
      GoDeep (* No subgoals left *)
    | Idtac (* This is the only subgoal left by this tactic *)
    ]
  ).

(*****************************)
(* Other three useful tactics. *)
(*****************************)

Tactic Definition IsNotANotLabel :=
 Unfold not;
 Induction 1;
 Intros;
 Discriminate.

Tactic Definition Apply_SAParallelPL :=
 Apply SAParallelPL;
  [ IsNotANotLabel
  | Idtac (* The only goal left *)
  ].

Tactic Definition Apply_SAParallelPR :=
 Apply SAParallelPR;
  [ IsNotANotLabel
  | Idtac (* The only goal left *)
  ].

Tactic Definition Apply_SAParallelNL :=
 Apply SAParallelNL;
  [ Idtac (* The only goal left *)
  | Repeat Constructor ;
    Unfold not;
    Intro;
    Discriminate
  ].

Tactic Definition Apply_SAParallelNR :=
 Apply SAParallelNR;
  [ Idtac (* The only goal left *)
  | Repeat Constructor ;
    Unfold not;
    Intro;
    Discriminate
  ].

(**************** End of tactic definitions *********************)

(*
Lemma example1_final_agent'_stuck :
*)
Axiom example1_final_agent'_stuck :
 (TransitionSystems.stuck ? step example1_final_agent').
(*
Proof.
 Unfold TransitionSystems.stuck;
 Unfold not;
 Intro;
 Elim H;
 Clear H;
 Intros;
 Inversion H;
  [ Do 5
     (Match Context With
       [H: (OStep ? ? ?) |- ?] -> Inversion H ; Clear H
     )
  | Idtac
  ].
Qed.
*)

Theorem example1:
 (reduces example1_initial_agent example1_final_agent').
Proof.
 Unfold example1_initial_agent example1_final_agent';
 Simpl;
 Unfold example1_program agent_of_program TheoryList_compl.fold_left_i;
 Unfold program_of_instruction;
 Simpl;
 Split;
  [ Unfold TransitionSystems.step_star;
    (*** First loop: 2+3 ==> 3+2 ***)
    (** Instruction 1 **)
    (* PC BangIn *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       EApply SAParallelTauIn1;
        [ Idtac (* We do not know yet *)
        | (* The program counter. Path: L *)
          Apply_SAParallelPL;
          Apply SMessage
        ];
       (* The instruction. Path: LLR *)
       Do 2 Apply_SAParallelPL;
       Apply_SAParallelPR;
       Constructor
     | Idtac (* This leaves an open goal *)
     ];
    (* (1) Inp (true) *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       EApply SAParallelTauIn1;
       [ Idtac (* We do not know yet *)
       | (* The register. Path: RRLL *)
         Do 2 Apply_SAParallelPR;
         Do 2 Apply_SAParallelPL;
         Apply SMessage
       ];
       (* The instruction. Path: LLRL *)
       Do 2 Apply_SAParallelPL;
       Apply_SAParallelPR;
       Apply_SAParallelPL;
       Constructor
     | Idtac
     ];
    (* PC emission and rendering. Path: LLLRL *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       Do 3 Apply_SAParallelPL;
       Apply_SAParallelPR;
       Apply_SAParallelPL;
       Constructor
     | Idtac
     ];



    (** Instruction 2 **)
    (* PC BangIn *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       (* Path common to both the PC and the instruction: LL *)
       Do 2 Apply_SAParallelPL;
       EApply SAParallelTauIn2;
        [ (* The program counter. Path: RLL *)
          Apply_SAParallelPR;
          Do 2 Apply_SAParallelPL;
          Apply SMessage
        | (* The instruction. Path: *)
          Constructor
        ]
     | Idtac (* This leaves an open goal *)
     ];
    (* (0) emission and rendering. Path: LLRL *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       Do 2 Apply_SAParallelPL;
       Apply_SAParallelPR;
       Apply_SAParallelPL;
       Constructor
     | Idtac
     ];
    (* PC emission and rendering. Path: LLRLR *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       Do 2 Apply_SAParallelPL;
       Apply_SAParallelPR;
       Apply_SAParallelPL;
       Apply_SAParallelPR;
       Constructor
     | Idtac
     ];



    (** Instruction 3 **)
    (* PC BangIn *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       (* Path common to both the PC and the instruction: L *)
       Apply_SAParallelPL;
       EApply SAParallelTauIn2;
        [ (* The program counter. Path: RLRL *)
          Apply_SAParallelPR;
          Apply_SAParallelPL;
          Apply_SAParallelPR;
          Apply_SAParallelPL;
          Apply SMessage
        | (* The instruction. Path: *)
          Constructor
        ]
     | Idtac (* This leaves an open goal *)
     ];
    (* (2) Inp (false) *)
    EApply Rstar_n;
     [ Unfold step;
       Right;
       Exists (2);
       (* The instruction. Path: LRL *)
       Apply_SAParallelNL;
       Apply_SAParallelNR;
       Apply_SAParallelNL;
       Constructor
     | Idtac
     ];
    (* PC emission and rendering. Path: LRL *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       Apply_SAParallelPL;
       Apply_SAParallelPR;
       Apply_SAParallelPL;
       Constructor
     | Idtac
     ];



    (*** Second loop: 3+2 ==> 4+1 ***)
    (** Instruction 1 **)
    (* PC BangIn *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       (* Path common to both the PC and the instruction: L *)
       Apply_SAParallelPL;
       EApply SAParallelTauIn1;
       [ Idtac (* We do not know yet *)
       | (* The register. Path: LL *)
         Do 2 Apply_SAParallelPL;
         Apply SMessage
       ];
       (* The instruction. Path: LRR *)
       Apply_SAParallelPL;
       Do 2 Apply_SAParallelPR;
       Constructor
     | Idtac (* This leaves an open goal *)
     ];
    (* (1) Inp (true) *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       EApply SAParallelTauIn1;
       [ Idtac (* We do not know yet *)
       | (* The register. Path: RRLRL *)
         Do 2 Apply_SAParallelPR;
         Apply_SAParallelPL;
         Apply_SAParallelPR;
         Apply_SAParallelPL;
         Apply SMessage
       ];
       (* The instruction. Path: LLRRL *)
       Do 2 Apply_SAParallelPL;
       Do 2 Apply_SAParallelPR;
       Apply_SAParallelPL;
       Constructor
     | Idtac
     ];
    (* PC emission and rendering. Path: LLLRRL *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       Do 3 Apply_SAParallelPL;
       Do 2 Apply_SAParallelPR;
       Apply_SAParallelPL;
       Constructor
     | Idtac
     ];



    (** Instruction 2 **)
    (* PC BangIn *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       (* Path common to both the PC and the instruction: LL *)
       Do 2 Apply_SAParallelPL;
       EApply SAParallelTauIn2;
        [ (* The program counter. Path: RRLL *)
          Do 2 Apply_SAParallelPR;
          Do 2 Apply_SAParallelPL;
          Apply SMessage
        | (* The instruction. Path: R *)
          Apply_SAParallelPR;
          Constructor
        ]
     | Idtac (* This leaves an open goal *)
     ];
    (* (0) emission and rendering. Path: LLRRL *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       Do 2 Apply_SAParallelPL;
       Do 2 Apply_SAParallelPR;
       Apply_SAParallelPL;
       Constructor
     | Idtac
     ];
    (* PC emission and rendering. Path: LLRRLR *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       Do 2 Apply_SAParallelPL;
       Do 2 Apply_SAParallelPR;
       Apply_SAParallelPL;
       Apply_SAParallelPR;
       Constructor
     | Idtac
     ];



    (** Instruction 3 **)
    (* PC BangIn *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       (* Path common to both the PC and the instruction: L *)
       Apply_SAParallelPL;
       EApply SAParallelTauIn2;
        [ (* The program counter. Path: RRLRL *)
          Do 2 Apply_SAParallelPR;
          Apply_SAParallelPL;
          Apply_SAParallelPR;
          Apply_SAParallelPL;
          Apply SMessage
        | (* The instruction. Path: R *)
          Apply_SAParallelPR;
          Constructor
        ]
     | Idtac (* This leaves an open goal *)
     ];
    (* (2) Inp (false) *)
    EApply Rstar_n;
     [ Unfold step;
       Right;
       Exists (2);
       (* The instruction. Path: LRRL *)
       Apply_SAParallelNL;
       Do 2 Apply_SAParallelNR;
       Apply_SAParallelNL;
       Constructor
     | Idtac
     ];
    (* PC emission and rendering. Path: LRRL *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       Apply_SAParallelPL;
       Do 2 Apply_SAParallelPR;
       Apply_SAParallelPL;
       Constructor
     | Idtac
     ];




    (*** Third loop: 4+1 ==> 5+0 ***)
    (** Instruction 1 **)
    (* PC BangIn *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       (* Path common to both the PC and the instruction: L *)
       Apply_SAParallelPL;
       EApply SAParallelTauIn1;
       [ Idtac (* We do not know yet *)
       | (* The register. Path: RLL *)
         Apply_SAParallelPR;
         Do 2 Apply_SAParallelPL;
         Apply SMessage
       ];
       (* The instruction. Path: LRRR *)
       Apply_SAParallelPL;
       Do 3 Apply_SAParallelPR;
       Constructor
     | Idtac (* This leaves an open goal *)
     ];
    (* (1) Inp (true) *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       EApply SAParallelTauIn1;
       [ Idtac (* We do not know yet *)
       | (* The register. Path: RRLRRL *)
         Do 2 Apply_SAParallelPR;
         Apply_SAParallelPL;
         Do 2 Apply_SAParallelPR;
         Apply_SAParallelPL;
         Apply SMessage
       ];
       (* The instruction. Path: LLRRRL *)
       Do 2 Apply_SAParallelPL;
       Do 3 Apply_SAParallelPR;
       Apply_SAParallelPL;
       Constructor
     | Idtac
     ];
    (* PC emission and rendering. Path: LLLRRRL *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       Do 3 Apply_SAParallelPL;
       Do 3 Apply_SAParallelPR;
       Apply_SAParallelPL;
       Constructor
     | Idtac
     ];



    (** Instruction 2 **)
    (* PC BangIn *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       (* Path common to both the PC and the instruction: LL *)
       Do 2 Apply_SAParallelPL;
       EApply SAParallelTauIn2;
        [ (* The program counter. Path: RRRLL *)
          Do 3 Apply_SAParallelPR;
          Do 2 Apply_SAParallelPL;
          Apply SMessage
        | (* The instruction. Path: RR *)
          Do 2 Apply_SAParallelPR;
          Constructor
        ]
     | Idtac (* This leaves an open goal *)
     ];
    (* (0) emission and rendering. Path: LLRRRL *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       Do 2 Apply_SAParallelPL;
       Do 3 Apply_SAParallelPR;
       Apply_SAParallelPL;
       Constructor
     | Idtac
     ];
    (* PC emission and rendering. Path: LLRRRLR *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       Do 2 Apply_SAParallelPL;
       Do 3 Apply_SAParallelPR;
       Apply_SAParallelPL;
       Apply_SAParallelPR;
       Constructor
     | Idtac
     ];



    (** Instruction 3 **)
    (* PC BangIn *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       (* Path common to both the PC and the instruction: L *)
       Apply_SAParallelPL;
       EApply SAParallelTauIn2;
        [ (* The program counter. Path: RRRLRL *)
          Do 3 Apply_SAParallelPR;
          Apply_SAParallelPL;
          Apply_SAParallelPR;
          Apply_SAParallelPL;
          Apply SMessage
        | (* The instruction. Path: RR *)
          Do 2 Apply_SAParallelPR;
          Constructor
        ]
     | Idtac (* This leaves an open goal *)
     ];
    (* (2) Inp (false) *)
    EApply Rstar_n;
     [ Unfold step;
       Right;
       Exists (2);
       (* The instruction. Path: LRRRL *)
       Apply_SAParallelNL;
       Do 3 Apply_SAParallelNR;
       Apply_SAParallelNL;
       Constructor
     | Idtac
     ];
    (* PC emission and rendering. Path: LRRRL *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       Apply_SAParallelPL;
       Do 3 Apply_SAParallelPR;
       Apply_SAParallelPL;
       Constructor
     | Idtac
     ];




    (*** Loop exit: 5+0 ==> stuck ***)
    (** Instruction 1 **)
    (* PC BangIn *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       (* Path common to both the PC and the instruction: LL *)
       Apply_SAParallelPL;
       EApply SAParallelTauIn1;
       [ Idtac (* We do not know yet *)
       | (* The register. Path: RRLL *)
         Do 2 Apply_SAParallelPR;
         Do 2 Apply_SAParallelPL;
         Apply SMessage
       ];
       (* The instruction. Path: LRRRR *)
       Apply_SAParallelPL;
       Do 4 Apply_SAParallelPR;
       Constructor
     | Idtac (* This leaves an open goal *)
     ];
    (* (1) Inp (false) *)
    EApply Rstar_n;
     [ Unfold step;
       Right;
       Exists (1);
       (* The instruction. Path: LLLRRRRL *)
       Do 3 Apply_SAParallelNL;
       Do 4 Apply_SAParallelNR;
       Apply_SAParallelNL;
       Constructor
     | Idtac
     ];
    (* PC emission and rendering. Path: LLLRRRRL *)
    EApply Rstar_n;
     [ Unfold step;
       Left;
       Do 3 Apply_SAParallelPL;
       Do 4 Apply_SAParallelPR;
       Apply_SAParallelPL;
       Constructor
     | Idtac
     ];




    (* Here we are! *)
    Constructor
  | Apply example1_final_agent'_stuck
  ].
Qed.
